@@ -13,13 +13,13 @@ main()
 	 * The OSFILE trap
 	 */
 
-	x [ 0x327d ] = 0xd4;
+	x [ 0x327d ] = 0x22;
 
 	/*
 	 * The OSFSC trap
 	 */
 
-	x [ 0x31b1 ] = 0xd7;
+	x [ 0x31b1 ] = 0x02;
 
 	lseek ( fd, 0, SEEK_SET );
 	write ( fd, x, 16384 );
