@@ -1,5 +1,5 @@
 /*
- * Copyright (c) James Fidell 1994.
+ * Copyright (c) James Fidell 1994, 1995, 1996.
  *
  * Permission to use, copy, modify, distribute, and sell this software
  * and its documentation for any purpose is hereby granted without fee,
@@ -33,9 +33,10 @@
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 
-#include "Patchlevel.h"
 #include "Config.h"
 #include "Beeb.h"
+#include "Ops.h"
+#include "opcodes.h"
 #include "6502.h"
 #include "Display.h"
 #include "ADC.h"
@@ -55,8 +56,8 @@
 #include "TubeUla.h"
 #include "SerialUla.h"
 #include "VideoUla.h"
-#include "Ops.h"
-#include "opcodes.h"
+#include "Options.h"
+#include "Sound.h"
 
 #ifdef	EMUL_FS
 #include "EFS.h"
@@ -67,10 +68,10 @@
  * Misc. forward declarations...
  */
 
-void				SaveSnapshot ();
+void				SaveSnapshot();
 void				RestoreSnapshot ( char* );
 void				Beeb();
-void				DisplayHelp();
+static void			HardwareReset();
 
 /*
  * This lot are just for convenience so that the real variables can
@@ -109,56 +110,24 @@ unsigned int		NMIAddress;
 unsigned int		IRQAddress;
 unsigned int		ResetAddress;
 
+/*
+ * Names of files to load which are passed in as command line options.
+ */
+
+char				*OsRomName = OS_ROM;
+char				*Rom15Name = LANG_ROM;
+#ifdef	XDFS
+char				*Rom14Name = XDFS_ROM;
+#endif
+char				*SnapshotName;
+
 byteval				SnapshotRequested = 0;
 byteval				QuitEmulator = 0;
-char				*SnapshotName;
+byteval				BreakKeypress = 0;
 
 byteval 			MaskableInterruptRequest = 0;
 static byteval 		ClockCyclesSoFar = 0;
 
-#define	CheckIRQ() \
-	if ( MaskableInterruptRequest && !IRQDisableFlag ) \
-	{ \
-		/* \
-		 * We have to explicitly stack the status register here, otherwise \
-		 * it will be pushed on the stack with the BRK flag set. \
-		 */ \
- \
-		StackPC(); \
-		GenerateIRQStatusRegister; \
-		StackByte ( StatusRegister ); \
-		SetIRQDisableFlag; \
-		SetProgramCounter ( IRQAddress ); \
- \
-		/* \
-		 * Have to add clock cycles ourselves, otherwise we end up with \
-		 * mutually recursive references to AddClockCycles and CheckIRQ. \
-		 * We can't actually cause an IRQ at this stage, though, because \
-		 * interrupts have just been disabled (above). \
-		 */ \
- \
-		ClockCyclesSoFar += 7; \
-		if ( ClockCyclesSoFar > 100 ) \
-	    { \
-			ViaClockUpdate ( ClockCyclesSoFar ); \
-			ClockCyclesSoFar = 0; \
-		} \
-		MaskableInterruptRequest--; \
-	}
-
-
-#define	AddClockCycles(n) \
-{ \
-	ClockCyclesSoFar += n; \
-	if ( ClockCyclesSoFar > 100 ) \
-	{ \
-		ViaClockUpdate ( ClockCyclesSoFar ); \
-		ClockCyclesSoFar = 0; \
-		CheckIRQ(); \
-	} \
-}
-
-
 /*
  * Debugging stuff
  */
@@ -173,213 +142,35 @@ unsigned char		DebugLevel = 0;
 static unsigned long	InstrCount [ 256 ];
 #endif
 
+
+#ifdef	FASTCLOCK
+
+#define	AddClockCycles(n) \
+{ \
+	ClockCyclesSoFar += n; \
+	if ( ClockCyclesSoFar > FASTCLOCK ) \
+	{ \
+		ViaClockUpdate ( ClockCyclesSoFar ); \
+		ClockCyclesSoFar = 0; \
+	} \
+}
+
+#else	/* FASTCLOCK */
+
+#define	AddClockCycles(n)	ViaClockUpdate(n)
+
+#endif	/* FASTCLOCK */
+
+
+
 int
 main ( int argc, char** argv )
 {
-	char		c;
-	byteval		RomNumber;
-	int			i = 1, j, next;
-	char		*OsRomName = OS_ROM;
-	char		*Rom15Name = LANG_ROM;
-	char		*RomName;
+#ifdef	COUNT_INSTRS
+	int			i;
+#endif
 
-	while ( --argc )
-	{
-		if ( *argv [ i ] == '-' )
-		{
-			next = 0;
-			j = 1;
-			while ( !next && (( c = argv [ i ][ j++ ] ) != '\0' ))
-			{
-				switch ( c )
-				{
-					case 'o' :					/* OS ROM name */
-						if ( argv [ i ][ j ] == '\0' )
-						{
-							if ( argc > 1 )
-							{
-								i++;
-								argc--;
-								OsRomName = argv [ i ];
-								next = 1;
-							}
-							else
-							{
-								fprintf ( stderr, "syntax error\nfilename " );
-								fprintf ( stderr, "should follow -o\n" );
-								exit ( 1 );
-							}
-						}
-						else
-						{
-							OsRomName = &argv[i][j];
-							next = 1;
-						}
-						break;
-
-					case 'p' :					/* Paged ROM */
-						switch ( argv[i][j] )
-						{
-							case '\0' :			/* No number, use 15 */
-								RomNumber = 15;
-								break;
-							case '0' : case '1' : case '2' :
-							case '3' : case '4' : case '5' :
-							case '6' : case '7' : case '8' :
-							case '9' :
-								RomNumber = atoi ( &argv[i][j] );
-								if ( RomNumber > 15 )
-								{
-									fprintf ( stderr, "ROM no. too big\n");
-									exit ( 1 );
-								}
-								break;
-							default :
-								fprintf ( stderr, "Bad ROM number for -p\n" );
-								fprintf ( stderr, "use -h for help\n" );
-								exit ( 1 );
-						}
-
-						if ( argc > 1 )
-						{
-							i++;
-							argc--;
-							RomName = argv [ i ];
-							next = 1;
-						}
-						else
-						{
-							fprintf ( stderr, "syntax error\nfilename " );
-							fprintf ( stderr, "should follow -p\n" );
-							exit ( 1 );
-						}
-
-						if ( RomNumber < 15 )
-							LoadPagedRom ( RomName, RomNumber );
-						else
-							Rom15Name = RomName;
-						break;
-
-					case 'w' :					/* Make a ROM slot writeable */
-						if ( argv [ i ][ j ] == '\0' )
-						{
-							if ( argc > 1 )
-							{
-								i++;
-								argc--;
-								next = 1;
-								if ( isdigit ( *argv [ i ] ))
-								{
-									RomNumber = atoi ( argv[i] );
-									if ( RomNumber < 16 )
-										PageWrite [ RomNumber ] = 1;
-									else
-									{
-										fprintf ( stderr, "ROM no. too big\n");
-										exit ( 1 );
-									}
-								}
-								break;
-							}
-
-							fprintf ( stderr, "syntax error\n rom number " );
-							fprintf ( stderr, "should follow -w\n" );
-							exit ( 1 );
-						}
-						else
-						{
-							if ( isdigit ( argv [ i ][ j ]))
-							{
-								RomNumber = atoi ( &argv[ i ][ j ]);
-								if ( RomNumber < 16 )
-									PageWrite [ RomNumber ] = 1;
-								else
-								{
-									fprintf ( stderr, "ROM no. too big\n");
-									exit ( 1 );
-								}
-							}
-							else
-							{
-								fprintf( stderr, "syntax error\n rom number ");
-								fprintf ( stderr, "should follow -w\n" );
-								exit ( 1 );
-							}
-						}
-						next = 1;
-						break;
-
-					case 'd' :					/* Disk directory/image */
-					{
-						char		*disk;
-
-						if ( argv [ i ][ j ] == '\0' )
-						{
-							if ( argc > 1 )
-							{
-								i++;
-								argc--;
-								disk = argv [ i ];
-								next = 1;
-								
-							}
-							else
-							{
-								fprintf ( stderr, "syntax error\nfilename " );
-								fprintf ( stderr, "should follow -d\n" );
-								exit ( 1 );
-							}
-						}
-						else
-						{
-							disk = &argv[i][j];
-							next = 1;
-						}
-						if ( ChangeDiskDirectory ( disk ))
-						{
-							fprintf ( stderr, "%s doesn't appear to ", disk );
-							fprintf ( stderr, "contain a valid disk image\n" );
-							exit ( 1 );
-						}
-						break;
-					}
-					case 'h' : case '?' :		/* Help message and exit */
-						DisplayHelp();
-						exit ( 1 );
-						break;
-
-					case 'V' :
-						fprintf ( stderr, "xbeeb %d.%d patchlevel %d\n",
-							VERSION, RELEASE, PATCHLEVEL );
-						exit ( 0 );
-						break;
-
-					default :					/* Unrecognised option */
-						fprintf ( stderr, "unrecognised option '%c'\n", c );
-						fprintf ( stderr, "use -h for help\n" );
-						exit ( 1 );
-						break;
-				}
-			}
-			i++;
-		}
-		else
-		{
-			/*
-			 * Assume a string on it's own is a snapshot name.
-			 * Fall over if we already have one.
-			 */
-
-			if ( SnapshotName )
-			{
-				fprintf ( stderr, "bad option '%s'\n", argv [ i ]);
-				fprintf ( stderr, "use -h for help\n" );
-				exit ( 1 );
-			}
-			else
-				SnapshotName = argv [ i ];
-		}
-	}
+	ProcessOptions ( argc, argv );
 
 	/*
 	 * Load the OS and initialise the CPU jump vectors.
@@ -395,9 +186,13 @@ main ( int argc, char** argv )
 	 */
 
 	LoadPagedRom ( Rom15Name, 15 );
+#ifdef	XDFS
+	LoadPagedRom ( Rom14Name, 14 );
+#endif
 
 	InitialiseKeyboard();
 	InitialiseScreen();
+	InitialiseSound();
 
 	Beeb();
 
@@ -414,21 +209,6 @@ main ( int argc, char** argv )
 }
 
 
-void
-DisplayHelp()
-{
-	printf ( "xbeeb -[Vh?] [<snapshot>] [-o <OS ROM>]" );
-	printf ( " [-p[0-15] <paged ROM>] [-w <num>]\n" );
-	printf ( "\t-[h?]\t\tdisplay this help message\n" );
-	printf ( "\t-V\t\tdisplay the version of xbeeb\n" );
-	printf ( "\t-p[<num>] <rom>\tload <rom> as ROM <num> (default 15)\n" );
-	printf ( "\t-o <rom>\tload <rom> as the OS\n" );
-	printf ( "\t-w <rom>\tmake <rom> writeable\n" );
-	printf ( "\n\n" );
-	return;
-}
-
-
 void
 Beeb()
 {
@@ -444,14 +224,14 @@ Beeb()
 
 	unsigned int			OpCount = 0;
 
-	InitProgramCounter = ResetAddress;
-
 	/*
 	 * Now load a snapshot if one was specified.
 	 */
 
 	if ( SnapshotName )
 		RestoreSnapshot ( SnapshotName );
+	else
+		HardwareReset();
 
 	Accumulator = InitAccumulator;
 	RegisterX = InitRegisterX;
@@ -471,6 +251,29 @@ Beeb()
 
 #include "6502.c"
 
+		/*
+		 * Now check to see if we need to service an interrupt
+		 */
+
+		if ( MaskableInterruptRequest && !IRQDisableFlag )
+		{
+			/*
+			 * We have to explicitly stack the status register here, otherwise
+			 * it will be pushed on the stack with the BRK flag set.
+			 */
+
+			StackPC();
+			GenerateIRQStatusRegister;
+			StackByte ( StatusRegister );
+			SetIRQDisableFlag;
+#ifndef	M6502
+			DecimalModeFlag = 0x0;
+#endif
+			SetProgramCounter ( IRQAddress );
+			MaskableInterruptRequest--;
+			AddClockCycles ( CLK_INTERRUPT );
+		}
+
 		if (( ++OpCount & 0x7ff ) == 0 )
 		{
 			CheckEvents();
@@ -495,6 +298,17 @@ Beeb()
 			else
 				if ( QuitEmulator )
 					break;
+				else
+					if ( BreakKeypress )
+					{
+						do
+						{
+							CheckEvents();
+						} while ( BreakKeypress );
+
+						HardwareReset();
+						SetProgramCounter ( InitProgramCounter );
+					}
 		}
 
 #ifdef	LIMIT
@@ -503,11 +317,52 @@ Beeb()
 	}
 #endif
 
+ExitEmulator:
+
 	ShutdownScreen();
 	return;
 }
 
 
+static void
+HardwareReset()
+{
+	/*
+	 * Do a power-up reset for all the hardware...
+	 */
+
+	ResetADConverter();
+	ResetAcia();
+	ResetCrtc();
+	ResetDiskController();
+	ResetEconetController();
+	ResetSerialUla();
+	ResetSystemVia();
+	ResetTubeUla();
+	ResetUserVia();
+	ResetVideoUla();
+
+	/*
+	 * Reset for the CPU is easy -- the only thing that gets initialised
+	 * is the program counter.
+	 */
+
+	InitProgramCounter = ResetAddress;
+
+#ifndef	M6502
+
+	/*
+	 * Except for the CMOS processors, which clear the D flag as well.
+	 */
+
+	DecimalModeFlag = 0x0;
+
+#endif	/* M6502 */
+
+	return;
+}
+
+
 #ifndef	AddClockCycles
 #ifdef	FASTCLOCK
 
@@ -515,7 +370,7 @@ void
 AddClockCycles ( byteval val )
 {
 	ClockCyclesSoFar += val;
-	if ( ClockCyclesSoFar > 100 )
+	if ( ClockCyclesSoFar > FASTCLOCK )
 	{
 		ViaClockUpdate ( ClockCyclesSoFar );
 		ClockCyclesSoFar = 0;
@@ -913,7 +768,7 @@ RestoreSnapshot ( char* sname )
 	 * various places in the hardware...
 	 */
 
-	RestoreColourMap();
+	return;
 }
 
 
