@@ -1,5 +1,5 @@
 /*
- * Copyright (c) James Fidell 1994.
+ * Copyright (c) James Fidell 1994, 1995, 1996.
  *
  * Permission to use, copy, modify, distribute, and sell this software
  * and its documentation for any purpose is hereby granted without fee,
@@ -26,10 +26,13 @@
 #ifndef	SERIALULA_H
 #define	SERIALULA_H
 
-extern  byteval		ReadSerialUla ( int );
-extern  void		WriteSerialUla ( int, byteval );
+extern  void			ResetSerialUla();
+extern  byteval			ReadSerialUla ( int );
+extern  void			WriteSerialUla ( int, byteval );
 
-extern	int			SaveSerialUla ( int );
-extern	int			RestoreSerialUla ( int, unsigned int );
+extern	int				SaveSerialUla ( int );
+extern	int				RestoreSerialUla ( int, unsigned int );
+
+extern	unsigned char	MotorLED;
 
 #endif	/* SERIALULA_H */
