@@ -1,5 +1,5 @@
 /*
- * Copyright (c) James Fidell 1994.
+ * Copyright (c) James Fidell 1994, 1995, 1996.
  *
  * Permission to use, copy, modify, distribute, and sell this software
  * and its documentation for any purpose is hereby granted without fee,
@@ -36,63 +36,153 @@
 #include "Sound.h"
 
 
-/*
- * FIX ME
- *
- * There ought to be a power-on state for all of the registers, but I
- * haven't got a clue what it is, so I write &00 to all of them apart from
- * IRB, which has the Speech unit INTERRUPT and READY signals hard-wired
- * high and the joystick button press lines hardwared as if the buttons
- * are not pressed.
- */
+static void				ClearPortAInterrupts ( void );
+static void				ClearPortBInterrupts ( void );
 
-Via		SystemVia =
+
+Via						SystemVia;
+int						SystemViaTimer1;
+int						SystemViaTimer2;
+byteval					SystemViaPortAPinLevel;
+byteval					SystemViaPortBPinLevel;
+unsigned char			SystemViaTimer2InterruptEnable;
+unsigned char			SystemViaTimer1Continuous;
+unsigned char			SystemViaTimer1SetPB7;
+unsigned char			SystemViaTimer2PulseCount;
+static unsigned char	PortALatchEnable;
+static unsigned char	PortBLatchEnable;
+static byteval			CA1Control, CA2Control, CB1Control, CB2Control;
+
+#define	DEFAULT_PORT_B	0xf0;
+
+
+void
+ResetSystemVia()
 {
-	0xf0, 0x00, 0x00, 0x00,
-	0x00, 0x00, 0x00, 0x00,
-	0x00, 0x00, 0x00, 0x00,
-	0x00, 0x00, 0x00, 0x00
-};
+	/*
+	 * The 6522 resets all registers to zero when it is powered up
+	 * or reset, except the counters and latches, which are random.
+	 * The SR should be random, too, but it always appears to be 0xff
+	 * on my machine.
+	 */
 
+	SystemViaPortAPinLevel = DEF_LOGIC;
+	SystemViaPortBPinLevel = DEF_LOGIC;
 
-int					SystemViaTimer1 = 0xffff;
-int					SystemViaTimer2 = 0xffff;
+	/*
+	 * FIX ME
+	 *
+	 * IRB has the Speech unit INTERRUPT and READY signals hard-wired
+	 * high and the joystick button press lines hardwared as if the buttons
+	 * are not pressed.
+	 */
 
-/*
- * FIX ME
- *
- * I'm not sure that Timer2InterruptEnable is needed.  In my docs. it
- * states that the T2 interrupt can only be re-enabled once triggered
- * by writing to T2CH or reading from T2CL.  It doesn't say if this is
- * because doing so resets the interrupt flag, or whether they act
- * independently.  I've assumed that they act independently because it
- * seems safer to do so.
- */
+	SystemVia [ IRB ] = DEFAULT_PORT_B;
+	SystemVia [ IRA ] = 0x00;
+	SystemVia [ ORB ] = 0x00;
+	SystemVia [ ORA ] = 0x00;
+	SystemVia [ DDRB ] = 0x00;
+	SystemVia [ DDRA ] = 0x00;
 
-unsigned char		SystemViaTimer2InterruptEnable = 0;
-unsigned char		SystemViaTimer1Continuous = 1;
-unsigned char		SystemViaTimer2PulseCount = 0;
+	/*
+	 * FIX ME
+	 *
+	 * Set these to random values...
+	 */
+
+	/*
+	 * This takes care of the counters...
+	 */
+
+	SystemViaTimer1 = 0x1234;
+	SystemViaTimer2 = 0x9876;
+
+	SystemVia [ T1LL ] = 0xaa;
+	SystemVia [ T1LH ] = 0x55;
+	SystemVia [ T2LL ] = 0xa5;
+
+	/*
+	 * FIX ME
+	 *
+	 * I'm not sure that Timer2InterruptEnable is needed.  In my docs. it
+	 * states that the T2 interrupt can only be re-enabled once triggered
+	 * by writing to T2CH or reading from T2CL.  It doesn't say if this is
+	 * because doing so resets the interrupt flag, or whether they act
+	 * independently.  I've assumed that they act independently because it
+	 * seems safer to do so.
+	 */
+
+	SystemViaTimer2InterruptEnable = 1;
+
+	/*
+	 * FIX ME
+	 *
+	 * This should really be 0x00.  However, the way the hardware is
+	 * designed, it looks like 0xff is a sensible default for this until
+	 * a more complete emulation is implemented.
+	 */
+
+	SystemVia [ SR ] = 0xff;
+
+	/*
+	 * FIX ME
+	 *
+	 * Need to handle the initial value of the SR control bits here, too.
+	 */
+
+	SystemVia [ ACR ] = 0x00;
+	SystemViaTimer1SetPB7 = 0;
+	SystemViaTimer1Continuous = 0;
+	SystemViaTimer2PulseCount = 0;
+	PortALatchEnable = 0;
+	PortBLatchEnable = 0;
+
+	SystemVia [ PCR ] = 0x0;
+	CA1Control = CA2Control = CB1Control = CB2Control = 0x0;
+
+	SystemVia [ IER ] = 0x0;
+	SystemVia [ IFR ] = 0x0;
+
+	/*
+	 * FIX ME
+	 *
+	 * Some values should be forced at the hardware now the register
+	 * values have been set -- write to the sound hardware for instance...
+	 */
+
+	return;
+}
 
-/*
- * This function is commented out because it's never used.
-static void			ClearPortAInterrupts ( void );
-*/
-static void			ClearPortBInterrupts ( void );
 
 byteval
 ReadSystemVia ( int addr )
 {
+	/*
+	 * Have to handle the address wrap-around -- the VIA only has 16
+	 * registers, but 32 memory locations are assigned to it.
+	 */
+
+	addr &= 0x0f;
 	switch ( addr )
 	{
-		case IRB :
+		case ORB :
 			/*
+			 * Returns the bit in the ORB when the corresponding DDRB bit
+			 * is set and either the pin level or the IRB bit when the
+			 * corresponding DDRB bit is clear, for non-latching/latching
+			 * modes respectively.
+			 *
 			 * In the default configuration, PB0-PB3 are outputs and
 			 * PB4-PB7 are inputs.
-			 *
+			 */
+
+			/*
 			 * Call ClearPortBInterrupts to clear the relevant flags
 			 * in the interrupt flag register.
 			 */
 
+			ClearPortBInterrupts();
+
 			/*
 			 * FIX ME
 			 *
@@ -107,29 +197,67 @@ ReadSystemVia ( int addr )
 			 * when the VIA is initialised above.
 			 */
 
-#ifdef	INFO
-			printf ( "read system VIA IRB, %2x\n", SystemVia [ IRB ] );
-#endif
 			ClearPortBInterrupts();
-			return ( SystemVia [ IRB ] );
+
+			if ( PortBLatchEnable )
+			{
+#ifdef	INFO
+			printf ( "read system VIA IRB = %02x\n", ( SystemVia [ ORB ] &
+				SystemVia [ DDRB ] ) | ( SystemVia [ IRB ] &
+				~SystemVia [ DDRB ] ));
+#endif
+				return (( SystemVia [ ORB ] & SystemVia [ DDRB ] ) |
+					( SystemVia [ IRB ] & ~SystemVia [ DDRB ] ));
+			}
+			else
+			{
+#ifdef	INFO
+			printf ( "read system VIA IRB = %02x\n", ( SystemVia [ ORB ] &
+				SystemVia [ DDRB ] ) | ( SystemViaPortBPinLevel &
+				~SystemVia [ DDRB ] ));
+#endif
+				return (( SystemVia [ ORB ] & SystemVia [ DDRB ] ) |
+					( SystemViaPortBPinLevel & ~SystemVia [ DDRB ] ));
+			}
 			break;
 
-		case IRA :
+		case ORA :
+		case ORA_nh :
 			/*
 			 * FIX ME
 			 *
-			 * Port A is used, but only through the non-handshaking
-			 * register.  If this one is being used then the hardware
-			 * is being screwed around with in some very strange
-			 * ways that I'm not sure how to handle, so I'll leave
-			 * this one for the time being.
-			 *
+			 * That AUG says that reading this should return the level on
+			 * the PA pin if latching is disabled or the IRA bit if latching
+			 * is enabled.  Seems strange that you can't read the ORA.  Can
+			 * that be correct ?
 			 */
 
-			fprintf ( stderr, "read from system VIA IRA" );
-			fprintf ( stderr, "Not yet implemented\n\n" );
-			ViaDump ( SystemVia );
-			FatalError();
+			/*
+			 * FIX ME
+			 *
+			 * Furthermore, I assume that reading the non-handshaking version
+			 * of the register always returns the pin levels.
+			 */
+
+
+			ClearPortAInterrupts();
+
+			if (( !PortBLatchEnable ) || ( addr == ORA_nh ))
+			{
+#ifdef  INFO
+				printf ( "read system VIA IRA(nh) = %02x\n",
+													SystemViaPortAPinLevel );
+#endif
+				return SystemViaPortAPinLevel;
+			}
+			else
+			{
+#ifdef	WARNINGS
+				fprintf ( stderr, "WARNING: read from system VIA IRA " );
+				fprintf ( stderr, "not yet implemented\n" );
+#endif
+				return SystemVia [ IRA ];
+			}
 			break;
 
 		/*
@@ -166,33 +294,28 @@ ReadSystemVia ( int addr )
 			break;
 
 		case T2CL :
-#ifdef	INFO
-			printf ( "read system VIA T2CL = 0x%2x\n", SystemVia [ T2CL ] );
-#endif
 			SystemViaTimer2InterruptEnable = 1;
-			return ( SystemVia [ T2CL ] );
+			return SystemViaTimer2 & 0xff;
 			break;
 
 		case T2CH :
-#ifdef	INFO
-			printf ( "read system VIA T2CH = 0x%2x\n", SystemVia [ T2CH ] );
-#endif
-			return ( SystemVia [ T2CH ] );
+			return SystemViaTimer2 >> 8;
 			break;
 
 		case SR :
 			/*
 			 * FIX ME
 			 *
-			 * This will need to do something once there's some serial
-			 * port emulation.
-			 *
+			 * This will need to do something sane once there's some
+			 * serial port emulation.  For the moment, just return a
+			 * random value.
 			 */
 
-			fprintf ( stderr, "read from system VIA SR\n" );
-			fprintf ( stderr, "Not yet implemented\n\n" );
-			ViaDump ( SystemVia );
-			FatalError();
+#ifdef	WARNINGS
+			fprintf ( stderr, "WARNING: read from system VIA SR " );
+			fprintf ( stderr, "not yet implemented\n" );
+#endif
+			return SystemVia [ SR ];
 			break;
 
 		case PCR :
@@ -222,100 +345,108 @@ ReadSystemVia ( int addr )
 			 * On reads bit 7 of the IER is always read as high.
 			 */
 #ifdef	INFO
-			printf ( "read system VIA IER = %2x\n", SystemVia [ IER ] | 0x80 );
+			printf ( "read system VIA IER = %02x\n", SystemVia [ IER ] | 0x80);
 #endif
 			return ( SystemVia [ IER ] | 0x80 );
 			break;
 
 		case IFR :
 #ifdef	INFO
-			printf ( "read system VIA IFR = %2x\n", SystemVia [ IFR ] );
+			printf ( "read system VIA IFR = %02x\n", SystemVia [ IFR ] );
 #endif
 			return ( SystemVia [ IFR ] );
 			break;
-
-		case IRA_nh :
-			/*
-			 * FIX ME
-			 *
-			 * Read/Write to ORA is supposed to clear the IRQ flag in the
-			 * IFR.  I don't think read/write to ORA_nh/IRA_nh is supposed
-			 * to do the same thing, but I'm not sure.
-			 *
-			 * This is an input from whichever device was last selected
-			 * using the 8-bit latch on port B.  I don't believe that
-			 * it's possible that the data could have changed since
-			 * then, but it's a potential problem.
-			 *
-			 */
-
-#ifdef	INFO
-			printf ( "read system VIA ORA(nh) = %2x\n", SystemVia [ ORA_nh ] );
-#endif
-			return ( SystemVia [ ORA_nh ] );
-			break;
-
-		default :
-			fprintf ( stderr, "out of range " );
-			fprintf ( stderr, "read from system VIA (addr = %x)\n", addr );
-			FatalError();
 	}
-	return 0;
+
+	/* NOTREACHED */
+
+	/*
+	 * FIX ME
+	 *
+	 * Should give a fatal error here ?
+	 */
+
+	return 0xff;
 }
 
 
 void
 WriteSystemVia ( int addr, byteval val )
 {
-	switch ( addr )
+	/*
+	 * Remember to handle the address wrap-around...
+	 */
+
+	switch ( addr & 0xf )
 	{
 		case ORB :
 		{
+			byteval			latch, enable;
+
 			/*
-			 * Only change the lines which are valid outputs...
+			 * The values in ORB always change, but the pin levels only
+			 * change as and when DDRB configures the pin as an output.
 			 */
 
-			byteval		old, new, ddr_in, ddr_out, latch, enable;
+			SystemVia [ ORB ] = val;
 
-			old = SystemVia [ ORB ];
-			ddr_in = ~SystemVia [ DDRB ];
-			ddr_out = SystemVia [ DDRB ];
+			val &= SystemVia [ DDRB ];
+			SystemViaPortBPinLevel &= ~SystemVia [ DDRB ];
+			SystemViaPortBPinLevel |= val;
 
-			new = ( old & ddr_in ) | ( val & ddr_out );
-			SystemVia [ ORB ] = new;
+			/*
+			 * FIX ME
+			 *
+			 * OK, so what do we do with the new values ?
+			 */
+
+			/*
+			 * Clear the interrupt flags
+			 */
+
+			ClearPortBInterrupts();
 #ifdef	INFO
-			printf ( "System VIA ORB = %2x\n", new );
+			printf ( "System VIA ORB = %02x, ", SystemVia [ ORB ] );
+			printf ( "Pin Levels = %02x\n", SystemViaPortBPinLevel );
 #endif
 
 			/*
-			 * We have an 8-bit latch addressed using ORB & 0x07, with
-			 * the new data value for the latch in ORB & 0x08.
-			 *
+			 * We have an 8-bit latch addressed using pin-levels & 0x07,
+			 * with the new data value for the latch in pin-levels & 0x08.
 			 */
 
-			latch = new & 0x7;
-			enable = new & 0x8;
+			latch = SystemViaPortBPinLevel & 0x7;
+			enable = SystemViaPortBPinLevel & 0x8;
 
 			switch ( latch )
 			{
 				case 0x0 :		/* Write enable for sound generator */
 					if ( !enable )
-						SoundWrite ( SystemVia [ ORA_nh ] );
+						SoundWrite ( SystemViaPortAPinLevel );
 					break;
 
 				case 0x1 :		/* Read enable for speech processor */
 					if ( !enable )
-						SpeechRead ( SystemVia[DDRA], &SystemVia [ ORA_nh ] );
+					{
+						SystemViaPortAPinLevel &= SystemVia [ DDRA ];
+						SystemViaPortAPinLevel |= ( SpeechRead() &
+							~SystemVia [ DDRA ] );
+					}
 					break;
 
 				case 0x2 :		/* Write enable for speech processor */
 					if ( !enable )
-						SpeechWrite ( SystemVia [ ORA_nh ] );
+						SpeechWrite ( SystemViaPortAPinLevel );
 					break;
 
 				case 0x3 :		/* Write enable for keyboard */
 					if ( !enable )
-						KeyboardWrite ( SystemVia[DDRA], &SystemVia[ORA_nh] );
+					{
+						SystemViaPortAPinLevel &= SystemVia [ DDRA ];
+						SystemViaPortAPinLevel |=
+							( KeyboardWrite ( SystemViaPortAPinLevel ) &
+							~SystemVia [ DDRA ] );
+					}
 					break;
 
 				case 0x4 :		/* Screen start for hardware scroll */
@@ -334,38 +465,113 @@ WriteSystemVia ( int addr, byteval val )
 					LedSetShiftLock ( enable );
 					break;
 			}
+			break;
+		}
+		case ORA :
+		case ORA_nh :
+		{
+			byteval			latch, enable;
+
+			/*
+			 * The values in ORA always change, but the pin levels only
+			 * change as and when DDRA configures the pin as an output.
+			 */
+
+			SystemVia [ ORA ] = val;
+
+			val &= SystemVia [ DDRA ];
+			SystemViaPortAPinLevel &= ~SystemVia [ DDRA ];
+			SystemViaPortAPinLevel |= val;
+
+#ifdef	INFO
+			if ( addr == ORA )
+				printf ( "System VIA ORA = %02x, ", SystemVia [ ORA ] );
+			else
+				printf ( "System VIA ORA(nh) = %02x, ", SystemVia [ ORA ] );
+			printf ( "Pin Levels = %02x\n", SystemViaPortAPinLevel );
+#endif
 
 			/*
 			 * Clear the interrupt flags
 			 */
 
-			ClearPortBInterrupts();
-			break;
-		}
-		case ORA :
+			ClearPortAInterrupts();
+
 			/*
-			 * FIX ME
+			 * The OS *always* uses ORA_nh as it's output port, so
+			 * if we're writing to ORA, barf here.  Otherwise, as I
+			 * understand it, there's no need to do anything with the
+			 * newly calculated values because the addressable latch
+			 * and it's PB3 input will need to be tweaked before anything
+			 * happens here.
 			 *
-			 * Don't know what do do about handling this at the moment.
-			 * Port A is used by the machine, but only through the non-
-			 * handshaking register.  Anything that deals with this
-			 * would be screwing with the hardware in some very strange
-			 * ways that I'm not sure how to handle.
+			 * There's always one, though, isn't there ?  So we take
+			 * account of the situation anyway...
+			 */
+
+			/*
+			 * We only have to deal with devices which respond to
+			 * writes on this port.
 			 *
 			 */
 
-			fprintf ( stderr, "write to system VIA ORA\n" );
-			fprintf ( stderr, "Not yet implemented\n\n" );
-			ViaDump ( SystemVia );
-			FatalError();
+			latch = SystemViaPortBPinLevel & 0x7;
+			enable = SystemViaPortBPinLevel & 0x8;
+
+			if ( !enable )
+			{
+				switch ( latch )
+				{
+					case 0x0 :		/* Write enable for sound generator */
+						SoundWrite ( SystemViaPortAPinLevel );
+						break;
+
+					case 0x2 :		/* Write enable for speech processor */
+						SpeechWrite ( SystemViaPortAPinLevel );
+						break;
+
+					case 0x3 :		/* Write enable for keyboard */
+						SystemViaPortAPinLevel &= SystemVia [ DDRA ];
+						SystemViaPortAPinLevel |=
+							( KeyboardWrite ( SystemViaPortAPinLevel ) &
+							~SystemVia [ DDRA ] );
+						break;
+				}
+			}
+			break;
+		}
+		case DDRB :
+			/*
+			 * FIX ME
+			 *
+			 * DEFAULT_PORT_B should at some stage be replaced by whatever
+			 * values are put on the input lines by the hardware we're
+			 * emulating as attached to user VIA port B.
+			 */
+
+			SystemVia [ DDRB ] = val;
+			SystemViaPortBPinLevel = ~val & DEFAULT_PORT_B;
+			SystemViaPortBPinLevel |= ( val & SystemVia [ ORB ] );
+#ifdef  INFO
+			printf ( "System VIA DDRB = %02x\n", val );
+#endif
+			return;
 			break;
 
-		case DDRB :
 		case DDRA :
-			SystemVia [ addr ] = val;
-#ifdef	INFO
-			printf ( "System VIA DDR%s = %2x\n",
-								( addr == DDRA ) ? "A" : "B", val );
+			/*
+			 * FIX ME
+			 *
+			 * DEF_LOGIC should at some stage be replaced by whatever
+			 * values are put on the input lines by the hardware we're
+			 * emulating as attached to user VIA port A.
+			 */
+
+			SystemVia [ DDRA ] = val;
+			SystemViaPortAPinLevel = ~val & DEF_LOGIC;
+			SystemViaPortAPinLevel |= ( val & SystemVia [ ORA ] );
+#ifdef  INFO
+			printf ( "System VIA DDRA = %02x\n", val );
 #endif
 			return;
 			break;
@@ -397,7 +603,11 @@ WriteSystemVia ( int addr, byteval val )
 			SystemVia [ T1CL ] = SystemVia [ T1LL ];
 			SystemViaTimer1 = SystemVia [ T1CH ] * 256 + SystemVia [ T1CL ];
 			SystemViaClearInterrupt ( INT_T1 );
-			WriteSystemVia ( ORB, ReadSystemVia ( ORB ) & 0x7f );
+			if ( SystemViaTimer1SetPB7 )
+			{
+				WriteSystemVia ( ORB, ReadSystemVia ( ORB ) & 0x7f );
+				SystemViaTimer1SetPB7 = 1;
+			}
 			break;
 
 		case T2CL :
@@ -450,25 +660,25 @@ WriteSystemVia ( int addr, byteval val )
 			 */
 
 			SystemVia [ ACR ] = val;
+			SystemViaTimer1SetPB7 = SystemVia [ ACR ] & 0x80;
 			SystemViaTimer1Continuous = SystemVia [ ACR ] & 0x40;
 			SystemViaTimer2PulseCount = SystemVia [ ACR ] & 0x20;
+			PortBLatchEnable = SystemVia [ ACR ] & 0x02;
+			PortALatchEnable = SystemVia [ ACR ] & 0x01;
 #ifdef	INFO
-			printf ( "System VIA ACR = %2x\n", SystemVia [ ACR ] );
+			printf ( "System VIA ACR = %02x\n", SystemVia [ ACR ] );
 #endif
 			break;
 
 		case PCR :
-			/*
-			 * FIX ME
-			 *
-			 * Controls what happens to CA1, CA2, CB1, CB2 and when
-			 * they might cause interrupts.
-			 *
-			 */
-
 			SystemVia [ PCR ] = val;
-#ifdef	INFO
-			fprintf( stderr, "WARNING: System VIA PCR = %2x\n",SystemVia[PCR]);
+			CA1Control = val & 0x01;
+			CA2Control = ( val >> 1 ) & 0x07;
+			CB1Control = ( val >> 4 ) & 0x01;
+			CB2Control = ( val >> 5 ) & 0x07;
+
+#ifdef	WARNINGS
+			fprintf(stderr, "WARNING: System VIA PCR = %02x\n",SystemVia[PCR]);
 #endif
 			break;
 
@@ -487,7 +697,7 @@ WriteSystemVia ( int addr, byteval val )
 			else
 				SystemVia [ IER ] &= ( val ^ 0x7f );
 #ifdef	INFO
-			printf ( "System VIA IER = %2x\n", SystemVia [ IER ] );
+			printf ( "System VIA IER = %02x\n", SystemVia [ IER ] );
 #endif
 
 			/*
@@ -500,74 +710,10 @@ WriteSystemVia ( int addr, byteval val )
 		case IFR :
 			SystemViaClearInterrupt ( val );
 #ifdef	INFO
-			printf ( "System VIA IFR = %2x\n", SystemVia [ IFR ] );
+			printf ( "System VIA IFR = %02x\n", SystemVia [ IFR ] );
 #endif
 			break;
 
-		case ORA_nh :
-		{
-			/*
-			 * Only change the lines which are valid outputs...
-			 */
-
-			byteval		old, new, ddr_in, ddr_out, latch, enable;
-
-			old = SystemVia [ addr ];
-			ddr_in = ~SystemVia [ DDRA ];
-			ddr_out = SystemVia [ DDRA ];
-
-			new = ( old & ddr_in ) | ( val & ddr_out );
-			SystemVia [ addr ] = new;
-
-#ifdef	INFO
-			printf ( "System VIA ORA(nh) = %2x\n", new );
-#endif
-			/*
-			 * Now the outputs have been set they should be picked up
-			 * by the correct device when the 3-bit latch in port B
-			 * is written.
-			 *
-			 * We only have to deal with devices which respond to
-			 * writes on this port, though.
-			 *
-			 */
-
-			latch = SystemVia [ ORB ] & 0x7;
-			enable = SystemVia [ ORB ] & 0x8;
-
-			switch ( latch )
-			{
-				case 0x0 :		/* Write enable for sound generator */
-					if ( !enable )
-						SoundWrite ( SystemVia [ ORA_nh ] );
-					break;
-
-				/*
-				 * Skip case 0x1 because it's for reading data into Port A,
-				 * not writing it...
-				 */
-
-				case 0x2 :		/* Write enable for speech processor */
-					if ( !enable )
-						SpeechWrite ( SystemVia [ ORA_nh ] );
-					break;
-
-				case 0x3 :		/* Write enable for keyboard */
-					if ( !enable )
-						KeyboardWrite ( SystemVia[DDRA], &SystemVia[ORA_nh] );
-					break;
-
-				default :
-					break;
-			}
-
-			break;
-		}
-
-		default :
-			fprintf ( stderr, "out of range " );
-			fprintf ( stderr, "write to system VIA (addr = %x)\n", addr );
-			FatalError();
 	}
 	return;
 }
@@ -616,7 +762,7 @@ SystemViaSetInterrupt( byteval IFR_flag )
 	 */
 
 #ifdef	INFO
-	printf ( "Interrupt requested by sys VIA, IFR flag 0x%x\n", IFR_flag );
+	printf ( "Interrupt requested by sys VIA, IFR flag 0x%02x\n", IFR_flag );
 #endif
 
 	if ( SystemVia [ IER ] & IFR_flag )
@@ -631,6 +777,44 @@ SystemViaSetInterrupt( byteval IFR_flag )
 }
 
 
+void
+SystemViaSetPortBPinLevel ( byteval val )
+{
+	SystemViaPortBPinLevel &= SystemVia [ DDRB ];
+	SystemViaPortBPinLevel |= ( val &= ~SystemVia [ DDRB ] );
+	return;
+}
+
+
+void
+SystemViaSetCA1 ( byteval level )
+{
+	if ( CA1Control == level )
+		SystemViaSetInterrupt ( INT_CA1 );
+	return;
+}
+
+
+void
+SystemViaSetCA2 ( byteval level )
+{
+	if ( !level && ( CA2Control == HS2_NEGATIVE || CA2Control ==
+														HS2_NEGATIVE_IND ))
+		SystemViaSetInterrupt ( INT_CA2 );
+	else
+		if ( level && ( CA2Control == HS2_POSITIVE || CA2Control ==
+														HS2_POSITIVE_IND ))
+		SystemViaSetInterrupt ( INT_CA2 );
+
+	/*
+	 * We don't have to worry about the other values of CA2Control -- they're
+	 * outputs.
+	 */
+
+	return;
+}
+
+
 static void
 ClearPortBInterrupts()
 {
@@ -638,26 +822,21 @@ ClearPortBInterrupts()
 	 * Always clear the CB1 active edge flag
 	 */
 
-	byteval		flags = 0x10;
-	byteval		pcr;
+	byteval		flags = INT_CB1;
 
 	/*
 	 * Clear the CB2 active edge flag unless the PCR indicates that
 	 * CB2 is in ``independent'' mode.
 	 */
 
-	pcr = SystemVia [ PCR ] & 0xe0;
-	flags &= (( pcr == 0x20 || pcr == 0x60 ) ? 0x08 : 0x0 );
+	if ( CB2Control == HS2_POSITIVE_IND || CB2Control == HS2_NEGATIVE_IND )
+		flags = INT_CB1 | INT_CB2;
 
 	SystemViaClearInterrupt ( flags );
 	return;
 }
 
 
-/*
- * This function is commented out because it's never used...
- */
-#if 0
 static void
 ClearPortAInterrupts()
 {
@@ -665,21 +844,20 @@ ClearPortAInterrupts()
 	 * Always clear the CA1 active edge flag
 	 */
 
-	byteval		flags = 0x02;
-	byteval		pcr;
+	byteval		flags = INT_CA1;
 
 	/*
 	 * Clear the CA2 active edge flag unless the PCR indicates that
 	 * CA2 is in ``independent'' mode.
 	 */
 
-	pcr = SystemVia [ PCR ] & 0x0e;
-	flags &= (( pcr == 0x02 || pcr == 0x06 ) ? 0x01 : 0x0 );
+	if ( CA2Control == HS2_POSITIVE_IND || CA2Control == HS2_NEGATIVE_IND )
+		flags = INT_CA1 | INT_CA2;
 
 	SystemViaClearInterrupt ( flags );
 	return;
 }
-#endif
+
 
 int
 SaveSystemVia ( int fd )
@@ -691,9 +869,11 @@ SaveSystemVia ( int fd )
 	SystemVia [ T2CL ] = SystemViaTimer2 & 0xff;
 	SystemVia [ T2CH ] = SystemViaTimer2 >> 8;
 
-	memcpy ( via, SystemVia, 17 );
+	memcpy ( via, SystemVia, 20 );
 	via [ 24 ] = SystemViaTimer2InterruptEnable;
 	via [ 25 ] = ScreenLengthIndex;
+	via [ 26 ] = SystemViaPortAPinLevel;
+	via [ 27 ] = SystemViaPortBPinLevel;
 
 	if ( write ( fd, via, 32 ) != 32 )
 		return -1;
@@ -710,14 +890,21 @@ RestoreSystemVia ( int fd, unsigned int ver )
 	if ( read ( fd, via, 32 ) != 32 )
 		return -1;
 
-	memcpy ( SystemVia, via, 17 );
+	memcpy ( SystemVia, via, 20 );
+
 	SystemViaTimer1 = SystemVia [ T1CL ] + ( SystemVia [ T1CH ] << 8 );
 	SystemViaTimer2 = SystemVia [ T2CL ] + ( SystemVia [ T2CH ] << 8 );
+
 	SystemViaTimer2InterruptEnable = via [ 24 ];
+	ScreenLengthIndex = via [ 25 ];
+	SystemViaPortAPinLevel = via [ 26 ];
+	SystemViaPortBPinLevel = via [ 27 ];
+
+	SystemViaTimer1SetPB7 = SystemVia [ ACR ] & 0x80;
 	SystemViaTimer1Continuous = SystemVia [ ACR ] & 0x40;
 	SystemViaTimer2PulseCount = SystemVia [ ACR ] & 0x20;
-
-	ScreenLengthIndex = via [ 25 ];
+	PortBLatchEnable = SystemVia [ ACR ] & 0x02;
+	PortALatchEnable = SystemVia [ ACR ] & 0x01;
 
 	return 0;
 }
